#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <poll.h>
#include <time.h>
#include <errno.h>
#include <termios.h>
#include <ctype.h>
#include <stdbool.h>

#define PORT "9999"  // the port users will be connecting to

// Creates a client socket on port and connects to ip
// Returns socket fd
int create_client(const char *ip, const char *port) {
	struct addrinfo hints;
	struct addrinfo *res, *rp;

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_INET;
	hints.ai_socktype = SOCK_STREAM;

	int s;
	if ((s = getaddrinfo(ip, port, &hints, &res)) != 0) {
		fprintf(stderr, "getaddrinfo: %s\n", gai_strerror(s));
		exit(EXIT_FAILURE);
	}

	int sockfd;

	for (rp = res; rp != NULL; rp = rp->ai_next) {
		sockfd = socket(rp->ai_family, rp->ai_socktype, rp->ai_protocol);
		if (sockfd == -1) {
			perror("socket");
			exit(EXIT_FAILURE);
		}
		
		if (connect(sockfd, rp->ai_addr, rp->ai_addrlen) == 0)
			break;		  /* Success */

		close(sockfd);
	}

	freeaddrinfo(res);

	if (rp == NULL) {
		fprintf(stderr, "connect failed\n");
		exit(EXIT_FAILURE);
	}

	return sockfd;
}

int main(void) {
	struct termios termios_p;
	if (tcgetattr(STDIN_FILENO, &termios_p) < 0) {
		fprintf(stderr, "failed to get attr: %d, %s\r\n", STDIN_FILENO, strerror(errno));
		exit(EXIT_FAILURE);
	}
	termios_p.c_lflag &= ~(ECHO | ICANON);
	if (tcsetattr(STDIN_FILENO, TCSANOW, &termios_p) < 0) {
		fprintf(stderr, "failed to set attr: %d, %s\r\n", STDIN_FILENO, strerror(errno));
		exit(EXIT_FAILURE);
	}
	int sockfd = create_client("127.0.0.1", "9999");
	printf("enter your name maybe: ");
	fflush(stdout);
	char name[BUFSIZ] = { 0, };
	ssize_t nn = 0;
	while ((name[nn] = getchar()) != EOF) {
		if (name[nn] == '\n') {
			putchar('\n');
			fflush(stdout);
			break;
		} else if (nn > 0 && name[nn] == 127) {
			printf("\b \b");
			fflush(stdout);
			nn--;
		} else {
			putchar(name[nn++]);
			fflush(stdout);
		}
	}
	name[nn++] = '\0';
//	write(sockfd, name, nn);
	struct pollfd pfd, psfd;
	pfd.fd = 0;
	pfd.events = POLLIN;
	psfd.fd = sockfd;
	psfd.events = POLLIN;
	char lineptr[BUFSIZ] = { 0, };
	ssize_t n = 0;
	printf("%s: ", name);
	fflush(stdout);
	const struct timespec tspec = { .tv_sec = 0, .tv_nsec = 16666666 };
	bool flush = false;
	while (true) {
		poll(&pfd, 1, 0);
		if (pfd.revents & POLLIN) {
			lineptr[n] = getchar();
			if (lineptr[n] == 127) {
				if (n == 0) {
					lineptr[n] = '\0';
				} else {
					printf("\b \b");
					lineptr[n--] = '\0';
					flush = true;
				}
			} else if (lineptr[n] == '\n') {
				lineptr[n++] = '\0';
				char out[nn+n+1];
				strcpy(out, name);
				out[nn-1] = ':';
				out[nn] = ' ';
				strcpy(&out[nn+1], lineptr);
				out[nn+n] = '\0';
				write(sockfd, out, nn+n+1);
				printf("\r%s: ", name);
				n = 0;
				lineptr[0] = '\0';
				flush = true;
			} else if (isprint(lineptr[n])) {
				putchar(lineptr[n]);
				lineptr[++n] = '\0';
				flush = true;
			}
		}
		poll(&psfd, 1, 0);
		if (psfd.revents & POLLIN) {
			char buf[BUFSIZ];
			read(sockfd, buf, BUFSIZ);
			printf("\r%s", buf);
			int spacer = n + 10 - strlen(buf);
			if (spacer > 0) {
				char space[spacer+1];
				memset(space, ' ', spacer);
				space[spacer] = '\0';
				printf("%s", space);
			}
			printf("\n%s: %s", name, lineptr);
			flush = true;
		}
		if (flush) {
			fflush(stdout);
			flush = false;
		}
		nanosleep(&tspec, NULL);
	}
}
